/*
 *  serial.c: Copyright 2003-2004 Phil Whineray
 *
 *  This file is part of x70talk.
 *
 *  x70talk is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  x70talk is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with x70talk; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Some bluetooth parts taken from gnokii, rfcomm, distributed under the GPL:
 *   Copyright (C) 1999, 2000 Hugh Blemings & Pavel Janik ml.
 *   Copyright (C) 2002       Marcel Holtmann <marcel@holtmann.org>
 *   Copyright (C) 2002-2003  Marcel Holtmann <marcel@holtmann.org>
 */

/*
 * implement a set of routines for reading and writing to phone - geared
 * towards AT command set e.g. ^M line termination added on writing,
 * removed on reading.
 */
#include "phone.h"

#undef DEBUG_READ_ALL  /* Character by character */
#undef DEBUG_WRITE_RAW

int serial_debug = 0;

#ifdef HAVE_SYS_POLL_H
#define POLLINFAIL POLLIN
#define POLLINWORK POLLIN
#else
/* Implement POLL using select - handy with winsock */
struct pollfd {
  int fd;
  int events;
};
#define POLLINFAIL 1
#define POLLINWORK 1
int poll(struct pollfd *pollfd, int i, int t)
{
  struct timeval tv;
  fd_set rfds;
  int retval;

  tv.tv_sec = 1;
  tv.tv_usec = /* t * 1000 */0;

  FD_ZERO(&rfds);
  FD_SET(pollfd->fd, &rfds);

  retval = select(1, &rfds, NULL, NULL, &tv);
  if (retval < 0) {
    perror("select");
#ifdef DEBUG_READ_ALL
    fprintf(stderr, "select error: %d\n", retval);
#endif
    return pollfd->events;
  }
  return retval;
}
#endif

#ifdef HAVE_WINSOCK2_H
typedef struct {
  char b[8];
} bdaddr_t;
#ifndef SOCK_STREAM
#define SOCK_STREAM 1
#endif

#ifndef HAVE_WS2BTH_H
typedef struct _SOCKADDR_BTH {
  USHORT addressFamily;
  bdaddr_t btAddr;
  GUID serviceClassId;
  ULONG port;
} SOCKADDR_BTH, *PSOCKADDR_BTH;
#define BTHPROTO_RFCOMM 0003
#define AF_BTH 32
#endif
#endif

#if defined(AF_BLUETOOTH)
/* Remove the dependency on -lbluetooth */
static int str2ba_local(const char *str, bdaddr_t *ba)
{
  int i;
  char ch;
  unsigned int b[6];

  memset(ba, 0, sizeof(*ba));
  i = sscanf(str, "%x:%x:%x:%x:%x:%x%c", b+0, b+1, b+2, b+3, b+4, b+5, &ch);
  if (i != 6)
    return SERIAL_BAD_PARAMETER;
  if ((b[0] | b[1] | b[2] | b[3] | b[4] | b[5]) > 0xff)
    return SERIAL_BAD_PARAMETER;

/*
 * Note: in the original (MacOS) sources the bytes were reversed to those
 * needed on x86: from which we conclude that there are network/host
 * ordering issues at play; hopefully the following should work for anyone.
 */
  if (htonl(1L) == 1L) { /* big endian */
    ba->b[0] = b[0];
    ba->b[1] = b[1];
    ba->b[2] = b[2];
    ba->b[3] = b[3];
    ba->b[4] = b[4];
    ba->b[5] = b[5];
  } else {
    ba->b[0] = b[5];
    ba->b[1] = b[4];
    ba->b[2] = b[3];
    ba->b[3] = b[2];
    ba->b[4] = b[1];
    ba->b[5] = b[0];
  }

  return 0;
}
#endif

#if defined(AF_BTH)
static int str2ba_local(const char *str, bdaddr_t *ba)
{
  int i;
  char ch;
  unsigned int b[6];

printf("Win2ba\n");
  memset(ba, 0, sizeof(*ba));
  i = sscanf(str, "%x:%x:%x:%x:%x:%x%c", b+0, b+1, b+2, b+3, b+4, b+5, &ch);
  if (i != 6)
    return SERIAL_BAD_PARAMETER;
  if ((b[0] | b[1] | b[2] | b[3] | b[4] | b[5]) > 0xff)
    return SERIAL_BAD_PARAMETER;

/*
 * Note: in the original (MacOS) sources the bytes were reversed to those
 * needed on x86: from which we conclude that there are network/host
 * ordering issues at play; hopefully the following should work for anyone.
 */
    ba->b[2] = b[0];
    ba->b[3] = b[1];
    ba->b[4] = b[2];
    ba->b[5] = b[3];
    ba->b[6] = b[4];
    ba->b[7] = b[5];

  return 0;
}
#endif

#if defined(AF_BLUETOOTH) || defined(AF_BTH)
#ifdef HAVE_WINSOCK2_H
static TCHAR *GetLastErrorMessage(DWORD last_error)
{
   static TCHAR errmsg[512];

   if (!FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, last_error,
                      0, errmsg, 511, NULL))
   {
      /* if we fail, call ourself to find out why and return that error */
      return (GetLastErrorMessage(GetLastError()));  
   }
  
   return errmsg;
}

static int rfcomm_open(const char *address)
{
   SOCKADDR_BTH sa;
   WORD wVersionRequested = 0x202;
   WSADATA m_data;
   SOCKET s;
   int retval;

fprintf(stderr, "WSAStartup\n");
   if (WSAStartup(wVersionRequested, &m_data) != 0) {
     return -1;
   }

fprintf(stderr, "WSA socket\n");
   s = socket(AF_BTH, SOCK_STREAM, BTHPROTO_RFCOMM); 

   if (s == INVALID_SOCKET) {
     const DWORD lastError = GetLastError();
     printf("Failed to get bluetooth socket! %s\n",
                 GetLastErrorMessage(lastError));
     return -1;
   }

fprintf(stderr, "WSA connect\n");
   sa.addressFamily = AF_BTH;
   sa.port = /*channel*/1 & 0xff;
   str2ba_local(address, (bdaddr_t *)&sa.btAddr);
   if (connect (s, (SOCKADDR *)&sa, sizeof(sa)) != 0) {
     const DWORD lastError = GetLastError();
     printf("Failed to get bluetooth socket! %s\n",
                 GetLastErrorMessage(lastError));
     return -1;
   }
  
   //WSACleanup();
   return s;
}

#else
static int rfcomm_open(const char *address)
{
  struct sockaddr_rc laddr, raddr;
  int i, sk;

  laddr.rc_family = AF_BLUETOOTH;
  bacpy(&laddr.rc_bdaddr, BDADDR_ANY);
  raddr.rc_family = AF_BLUETOOTH;
  raddr.rc_channel = 1;
  i = str2ba_local(address, &raddr.rc_bdaddr);
  if (i)
    return i;

  if ((sk = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
    return sk;
  }

  if ((i = connect(sk, (struct sockaddr *)&raddr, sizeof(raddr))) < 0) {
    close(sk);
    return i;
  }

  return sk;
}
#endif

#endif

static int is_bluetooth_addr(const char *str)
{
  int i;
  unsigned int b[6];
  char ch;

  i = sscanf(str, "%x:%x:%x:%x:%x:%x%c", b+0, b+1, b+2, b+3, b+4, b+5, &ch);
  if (i != 6)
    return 0;
  return 1;
}

int serial_open(const char *path)
{
  int fd;

  if (is_bluetooth_addr(path)) {
#if defined(AF_BLUETOOTH) || defined(AF_BTH)
    fd = rfcomm_open(path);
#else
    errno = ENOSYS;
    return -1;
#endif
  } else {
#ifndef O_BINARY
#define O_BINARY 0
#endif
    fd = open(path, O_RDWR | O_BINARY);
    if (serial_debug)
      fprintf(stderr, "<o> <%s> = %d\n", path, fd);
  }
  return fd;
}

void serial_close(int fd)
{
  close(fd);
}

/*
 * Try to read anything that might be left on the line; wait for up to
 * 0.1 secs between reads, and give up if no data or we have read more than
 * 10 times the timeout limit.
 */
int serial_clear(int fd)
{
  struct pollfd pollfd;
  int flags, i;
  char tmpbuf[1];

  pollfd.fd = fd;
  pollfd.events = POLLINFAIL;
  for (i = 0; i < 500 && poll(&pollfd, 1, 100); i++) {
#ifdef DEBUG_READ_ALL
    fprintf(stderr, "Polled ok - now reading\n");
#endif
    read(fd, tmpbuf, sizeof(tmpbuf));
  }

  if (poll(&pollfd, 1, 100))
    return SERIAL_TIMEOUT;
  return 0;
}

int serial_write_raw(int fd, const char *buf, int len)
{
  int actuallen;

#ifdef DEBUG_WRITE_RAW
  {
    const char *c = buf;
    fprintf(stderr, "<W> <");
    while (c < buf+len)
      if (isprint(*c))
        fprintf(stderr, "%c", *c++);
      else
        fprintf(stderr, "[0x%x]", *c++);
    fprintf(stderr, ">\n");
  }
#endif
  actuallen = write(fd, buf, len);
  if (actuallen < 0)
    return actuallen;
  if (actuallen < len)
    return SERIAL_SHORT_WRITE;
  return 0;
}

int serial_write(int fd, const char *buf)
{
  int len = strlen(buf);
  int actuallen;

  if (serial_debug)
    fprintf(stderr, "<w> <%s>\n", buf);
  actuallen = write(fd, buf, strlen(buf));
  if (actuallen < 0)
    return actuallen;
  if (actuallen < len)
    return SERIAL_SHORT_WRITE;

  actuallen = write(fd, "\r\n", 2);
  if (actuallen < 0)
    return actuallen;
  return 0;
}

int serial_flush(int fd, int character, int num)
{
  struct pollfd pollfd;
  int i, count;
  char buf[4];

  if (serial_debug)
    fprintf(stderr, "Writing flush 0x%x * %d\n", character, num);
  buf[0] = character;
  buf[1] = 0;
  for (count = 0; count < num; count++) {
    i = write(fd, buf , 1);
    if (i < 0)
      return i;
    if (i == 0)
      return SERIAL_SHORT_WRITE;
  }

  if (serial_debug)
    fprintf(stderr, "Reading flush 0x%x * %d\n", character, num);
  pollfd.fd = fd;
  pollfd.events = POLLINFAIL;
  if (!poll(&pollfd, 1, TIMEOUT * 1000)) {
    if (serial_debug)
      fprintf(stderr, "Reading timeout\n");
    return SERIAL_TIMEOUT;
  }
  
  i = read(fd, buf, 1);
  while (i >= 0) {
    if (*buf == character) {
      num--;
      if (!num)
        return 0;
      if (serial_debug)
        fprintf(stderr, "Reading flush 0x%x * %d\n", character, num);
    } else {
      if (serial_debug)
        fprintf(stderr, "Reading flush %c\n", isprint(character)?character:'^');
    }
    if (!poll(&pollfd, 1, TIMEOUT * 1000)) {
      if (serial_debug)
        fprintf(stderr, "Reading timeout\n");
      return SERIAL_TIMEOUT;
    }
    i = read(fd, buf, 1);
  }
  return i;
}

/*
 * Read a single character. Ignore "\r".
 */
int serial_read_char(int fd, char *buf)
{
  struct pollfd pollfd;
  int i;

  pollfd.fd = fd;
  pollfd.events = POLLINWORK;
  if (!poll(&pollfd, 1, TIMEOUT * 1000)) {
    if (serial_debug)
      fprintf(stderr, "Reading timeout\n");
    return SERIAL_TIMEOUT;
  }
  i = read(fd, buf, 1);
#ifdef DEBUG_READ_ALL
  if (isprint(*buf))
      fprintf(stderr, "read %c\n", *buf);
  else
      fprintf(stderr, "read nonprint %d\n", *buf);
#endif
  if (i < 0)
    return i;
  return 0;
}

int serial_read(int fd, char *buf, int buflen)
{
  int i;
  char *ptr;
  char *end;

  ptr = buf;
  end = buf+buflen-1;

  i = serial_read_char(fd, ptr);
  while (i == 0) {
    if (*ptr == '\r') {
      i = serial_read_char(fd, ptr);
      continue;
    }

    if (*ptr == '\n') {
      *ptr = 0;
      if (serial_debug)
        fprintf(stderr, "<r> <%s>\n", buf);
      return 0;
    }

    if (*ptr == 032) {
      *ptr = 0;
      if (serial_debug)
        fprintf(stderr, "<r> <%s>\n", buf);
      return 0;
    }

    if (!isprint(*ptr)) {
      i = serial_read_char(fd, ptr);
      continue;
    }

    ptr++;
    if (ptr == end) {
#ifdef DEBUG_READ_ALL
      fprintf(stderr, "read maxbuf\n");
#endif
      *ptr = 0;
      if (serial_debug)
        fprintf(stderr, "<m> <%s>\n", buf);
      return 0;
    }

    i = serial_read_char(fd, ptr);
  }
  return i;
}

static int atdiscard(const char *buf, const char *command)
{
  static struct {
    char *text;
    int match_beginning;
  } ignores[] = {
    { "+CLIP:", 1 },
    { "RING", 0 },
    { "", 0 }
  };
  int i;
  int match;

  if (strcmp(buf, command) == 0)
    return 1;

  for (i = 0; i < sizeof(ignores)/sizeof(ignores[0]); i++) {
    if (ignores[i].match_beginning)
      match = strncmp(buf, ignores[i].text, strlen(ignores[i].text));
    else
      match = strcmp(buf, ignores[i].text);
    if (match == 0)
      return 1;
  }

  return 0;
}

static int aterror(const char *buf)
{
  static struct {
    char *text;
    int match_beginning;
    int error;
  } errors[] = {
    { "+CME ERROR: 22", 0, 0 }, /* Empty phonebook: pass up */
    { "+CME ERROR: 14", 0, 0 }, /* No signal: pass up */
    { "+CME ERROR:", 1, AT_ERROR },
    { "ERROR", 0, AT_ERROR },
    { "OK", 0, 0 }
  };
  int i;
  int match;

  for (i = 0; i < sizeof(errors)/sizeof(errors[0]); i++) {
    if (errors[i].match_beginning)
      match = strncmp(buf, errors[i].text, strlen(errors[i].text));
    else
      match = strcmp(buf, errors[i].text);
    if (match == 0)
      return errors[i].error;
  }

  return -1;
}

static int atcmd(int fd, const char *command, int len, int eol,
                 struct at_response_list **response)
{
  struct at_response_list *resp = NULL, *resp_tail = NULL;
  char buf[sizeof(resp->line)], *end, *c;
  int i, err;

  *buf = eol;
  *response = NULL;

  if (serial_debug) {
     const char *x = command, *end = x+len;
     fprintf(stderr, "<c> <", buf);
     while (x < end) {
       fputc(*x?*x:'@', stderr);
       x++;
     }
     fprintf(stderr, ">\n");
     fflush(stderr);
  }

  serial_write_raw(fd, command, len);
  if (eol) {
    if (eol == '\n')
      serial_write_raw(fd, "\r\n", 2);
    else
      serial_write_raw(fd, buf, 1);
  }

  /* Start reading from phone */
  c = buf;
  end = buf+sizeof(buf)-1;
  i = serial_read_char(fd, c);
  while (c < end && i == 0) {
    if (*c == '\0') { /* Dubious unicode handling... */
      i = serial_read_char(fd, c);
      continue;
    }

    if (*c == ' ' && c == buf+1 && *buf == '>')
      return AT_MORE;

    if (*c == '\r') {
      i = serial_read_char(fd, c);
      continue;
    }

    if (*c == '\n' || *c == 032) {
      eol = *c;
      *c = 0;

      if (atdiscard(buf, command)) {
        if (serial_debug && *buf)
          fprintf(stderr, "<r> <%s> discarded\n", buf);
        c = buf;
        i = serial_read_char(fd, c);
        continue;
      }

      if ((err = aterror(buf)) >= 0) {
        if (serial_debug)
          fprintf(stderr, "<r> <%s> error code %d\n", buf, err);
        return err;
      }

      if (serial_debug)
        fprintf(stderr, "<r> <%s> stored\n", buf);
/* Allocate a response structure */
      resp = malloc(sizeof(struct at_response_list));
      if (!resp)
        return AT_MEMORY_ERROR;
      memset(resp, 0, sizeof(resp));
      strcpy(resp->line, buf);
      resp->eol_char = eol;
      if (resp_tail) {
        resp_tail->next = resp;
        resp_tail = resp;
      } else {
        *response = resp_tail = resp;
      }
      c = buf;
      i = serial_read_char(fd, c);
      continue;
    }
    i = serial_read_char(fd, ++c);
  }

  if (c == end)
    return AT_RESPONSE_TOO_BIG;
  return i;
}

int at_run(int fd, const char *command, int len,
           struct at_response_list **response)
{
  return atcmd(fd, command, len?len:strlen(command), '\n', response);
}

int at_more(int fd, const char *command, int len,
            struct at_response_list **response)
{
  return atcmd(fd, command, len?len:strlen(command), 032, response);
}

void at_free(struct at_response_list **response)
{
  struct at_response_list *head = *response, *next;
  while (head) {
    next = head->next;
    free(head);
    head = next;
  }
  *response = NULL;
}
